import os

import pytest

from ticket_cross_check.gitlab_connector import GitlabConnector
from ticket_cross_check.gitlab_models import GitlabIssue


def need_token(func):
    @pytest.mark.skipif(os.getenv('PRIVATE_TOKEN') is None, reason="Need gitlab private token %s" % os.environ)
    def myfunc(*args, **kwargs):
        return func(*args, **kwargs)

    return myfunc


def need_tcc_as_project(func):
    """This thest would need the ticket cross-check gilab project to be accessible during test"""
    TCC_PID = 38031020

    @pytest.mark.skipif(os.getenv('PROJECT_ID') is TCC_PID or os.getenv('CI_PROJECT_ID') is TCC_PID,
                        reason="Need access to TCC project")
    def myfunc(*args, **kwargs):
        return func(*args, **kwargs)

    return myfunc


def gitlab_connector():
    try:
        return GitlabConnector.factory()
    finally:
        pass


@need_token
def test_init_from_env():
    gc = GitlabConnector(personal_api_token=os.getenv('PRIVATE_TOKEN'), project_id=os.getenv('PROJECT_ID'))
    assert isinstance(gc, GitlabConnector)


@need_token
def test_factory():
    assert isinstance(gitlab_connector(), GitlabConnector)


@need_token
def test_get_issues():
    result = gitlab_connector().get_issues()
    assert result
    assert isinstance(result, set)
    assert isinstance(result.pop(), GitlabIssue)


@need_token
@need_tcc_as_project
def test_issue_labels_exist_in_gitlab_project():
    result = gitlab_connector().get_project().labels.list()
    labelnames = [label.name for label in result]
    assert 'ignore' in labelnames
    assert 'ignore_doc' in labelnames


@need_token
@need_tcc_as_project
def test_isses_with_labels_exist():
    ignore_issue = gitlab_connector().get_project().issues.list(labels=['ignore'])
    assert 1 == len(ignore_issue)
    assert 'ignore' in ignore_issue[0].labels
    assert ['ignore'] == ignore_issue[0].labels



import json
import os.path
from pathlib import Path

from loguru import logger

from ticket_cross_check.gitlab_models import GitlabIssue
from .models import IssueFileMatch


def _get_unmatched_issues_from_files(issues_in_files: dict[int, list[IssueFileMatch]], issues: set[GitlabIssue]) \
        -> list[IssueFileMatch]:
    unmatched = []
    iids = [int(issue.iid) for issue in issues]
    for iid in issues_in_files.keys():
        logger.trace(iid)
        if iid not in iids:
            logger.debug(issues_in_files[iid])
            unmatched += issues_in_files[iid]
    return unmatched


def _get_un_and_solved_issued(
        issues_in_files: dict[int, list[IssueFileMatch]],
        issues: set[GitlabIssue],
        ignore_labels: set[str]
) -> tuple[set[GitlabIssue], set[IssueFileMatch]]:
    """

    :param issues_in_files:
    :param issues:
    :return: unsolved_issue set, solved_issue set
    """
    unsolved = set()
    solved = set()

    for issue in issues:
        if issue.ignore():
            continue
        if issue.iid not in issues_in_files:
            # do not mind if we shall ignore the issue anyway
            if set(issue.labels).intersection(ignore_labels):
                continue
            # otherwise make an entry
            unsolved.add(issue)
        else:
            for iif in issues_in_files[issue.iid]:
                iif.matched_issue = issue
                solved.add(iif)
    return unsolved, solved


def _get_label_set(search_dir: str) -> set:
    ls = set()
    ls.add('ignore_%s' % search_dir)
    bn = os.path.basename(search_dir)
    if bn != search_dir:
        ls.add('ignore_%s' % bn)
    return ls


def analyze_issues_vs_code(issues_in_files: dict[int, list[IssueFileMatch]], issues: set[GitlabIssue], search_dir: str):
    # Unsolved Requirement: issues not in files
    # Problem: file-issues without real issues
    # Solved: issue in file(s)
    ignore_lables = _get_label_set(search_dir)
    unsolved, solved = _get_un_and_solved_issued(issues_in_files, issues, ignore_lables)
    problems = _get_unmatched_issues_from_files(issues_in_files, issues)
    return problems, unsolved, solved


def write_as_json2file(path: Path, data: dict):
    with path.open(mode='w') as ofile:
        ofile.write(json.dumps(data, indent=2))


def render(
        solved: set[IssueFileMatch],
        unsolved: set[GitlabIssue],
        problems: list[IssueFileMatch],
        base_link: str,
        search_dir: str = ""
):
    data = []
    for ifm in solved:
        data.append([
            ifm.iid,
            f"<a href='{ifm.get_issue_link()}'>{ifm.matched_issue.title}</a>",
            f"<a href='{ifm.get_deep_link('main')}'>{ifm.filename}</a>",
            ifm.line_nr,
            "solved",
            search_dir
        ])

    status = "unsolved"
    for issue in unsolved:
        data.append([
            issue.iid,
            f"<a href='{issue.get_issue_link()}'>{issue.title}</a>",
            '',
            '',
            status,
            search_dir
        ])

    status = "problem"
    _id = "<span style='color: red'>%s</a>"
    for ifm in problems:
        data.append([
            _id % ifm.iid,
            '',
            f"<a href='{ifm.get_deep_link('main', base_link)}'>{ifm.filename}</a>",
            ifm.line_nr,
            status,
            search_dir
        ])

    return data
    # result = {
    #     'data': data,
    # }
    # _write_json2file(path, result)


def _get_issue_ignore_list(issue: GitlabIssue, dirs: list):
    """return binary list of dirs to ignore"""
    ilist = [0] * len(dirs)
    for label in issue.labels:
        if label.startswith('ignore_'):
            label_dir = label.replace('ignore_', '')
            # find dirs OR parts of them that match
            indexes = []
            if label_dir in dirs:
                indexes.append(dirs.index(label_dir))
            for dir in dirs:
                # search for basenames
                bn = os.path.basename(dir)
                if label_dir in bn:
                    indexes.append(dirs.index(dir))
            for idx in indexes:
                ilist[idx] = 1
    return ilist


def generate_base_matrix(gitlab_issues: set[GitlabIssue], dirs: list):
    matrix = {'_subdirs': dirs}
    for issue in gitlab_issues:
        if issue.ignore():
            continue
        iil = _get_issue_ignore_list(issue, dirs)
        matrix[issue.iid] = {'issue': issue, 'iid': issue.iid, 'counts': [0] * len(dirs), 'ignore': iil}
    return matrix


def build_issue_matrix(
        matrix_in: dict,
        solved: set[IssueFileMatch],
        search_dir: str = ""
):
    """
    increments "counts" to each matrix[issue_is] entry for every solved file
    """
    matrix = matrix_in.copy()
    count_index = matrix['_subdirs'].index(search_dir)
    for ifm in solved:
        issue_id = ifm.matched_issue.iid
        matrix[issue_id]['counts'][count_index] += 1
    return matrix


def render_matrix(matrix: dict):
    data = []
    issuelink = None
    for issue, value in matrix.items():
        if isinstance(issue, str):
            continue
        if not issuelink:
            end = value['issue'].get_issue_link().rfind('/')
            issuelink = value['issue'].get_issue_link()[:end]
        data.append(
            [
                issue,
                value['issue'].title,
            ] + value['counts'] + value['ignore']
        )

    return {
        "title": matrix['_subdirs'],
        "issuelink": issuelink,
        "data": data
    }
